from datetime import datetime
from turbogears.database import PackageHub
from sqlobject import *
from turbogears import identity
from problem_report import ProblemReport
import uuid

#import logging
#log = logging.getLogger("tgcrashdb.models")

hub = PackageHub('tgcrashdb')
__connection__ = hub

def random_uuid():
    return str(uuid.uuid4())

class CrashReport(SQLObject):
    '''Represents a crash report as emitted by apport.'''
    # UUID for this crash (so the user can look it up later)
    uuid = StringCol(alternateID=True,default=random_uuid,length=36)
    # Timestamp of update (so we know when a crash is out of date)
    timestamp = DateTimeCol(default=datetime.now)
    # Base ProblemReport items (from problem_report)
    ProblemType = EnumCol(enumValues=["Crash","Packaging","Kernel"],default="Crash")
    Date        = DateTimeCol(default=datetime.now)
    # Basic info about the crash (from apport.report.add_os_info())
    DistroRelease = UnicodeCol(default="")
    Architecture  = UnicodeCol(default="")
    Uname         = UnicodeCol(default="")
    NonfreeKernelModules = UnicodeCol(default="")
    # Info about the affected package (from apport.report.add_package_info())
    Package             = UnicodeCol(default="")
    SourcePackage       = UnicodeCol(default="")
    PackageArchitecture = UnicodeCol(default="")
    Dependencies        = UnicodeCol(default="")
    # /proc info (from apport.report.add_proc_info())
    ExecutablePath   = UnicodeCol(default="")
    InterpreterPath  = UnicodeCol(default="")
    ProcEnviron      = UnicodeCol(default="")
    ProcCmdline      = UnicodeCol(default="")
    ProcStatus       = UnicodeCol(default="")
    ProcMaps         = UnicodeCol(default="")
    # Core dump info (from the apport crash handler itself)
    Signal           = IntCol(default=0)
    # Set length=2**24 to force LONGBLOB
    #CoreDump         = PickleCol(default="",length=2**24) # might be empty
    # debugging info (from apport.report.add_gdb_info())
    Registers        = UnicodeCol(default="")
    Disassembly      = UnicodeCol(default="")
    Stacktrace       = UnicodeCol(default="")
    ThreadStacktrace = UnicodeCol(default="")
    StacktraceTop    = UnicodeCol(default="")
    # Bonus info
    Title            = UnicodeCol(default="")
    # Link to a bug report
    # FIXME causes a ValueError because it has no default... uhh.. what?
    bug = ForeignKey('BugLink', default=1)

    ## argle. base64-encode the coredump to work around !#@^!@& unicode errors
    #def _set_core_dump(self, value):
    #    self._SO_set_core_dump(value.encode('base64'))
    #
    #def _get_core_dump(self, value):
    #    return self._SO_get_core_dump().decode('base64')

    def readReport(self,r):
        for key,val in r.iteritems():
            if hasattr(self,key):
                # Handle a couple of special fields
                if key == "Date":
                    val = datetime.strptime(val,"%c")
                elif key == "Signal":
                    val = int(val)
                elif isinstance(val,str): # XXX fixes encoding problems?
                    val = val.decode('iso8859-15').encode('utf-8')
                setattr(self,key,val)

    def read(self,reportdata,binary=True):
        '''Read data in from a crash report file'''
        r = ProblemReport()
        r.load(reportdata,binary)
        self.readReport(r)

class BugLink(SQLObject):
    '''Represents a link to a bug report DB.'''
    # decides which set of methods to use
    type   = StringCol(default="bugzilla") 
    url    = StringCol(default="https://bugzilla.redhat.com/bugzilla/")
    bugid  = IntCol(default=0)
    crashreport = MultipleJoin('CrashReport', joinColumn='bug_id')

# identity models.
class Visit(SQLObject):
    """
    A visit to your site
    """
    class sqlmeta:
        table = 'visit'

    visit_key = StringCol(length=40, alternateID=True,
                          alternateMethodName='by_visit_key')
    created = DateTimeCol(default=datetime.now)
    expiry = DateTimeCol()

    def lookup_visit(cls, visit_key):
        try:
            return cls.by_visit_key(visit_key)
        except SQLObjectNotFound:
            return None
    lookup_visit = classmethod(lookup_visit)


class VisitIdentity(SQLObject):
    """
    A Visit that is link to a User object
    """
    visit_key = StringCol(length=40, alternateID=True,
                          alternateMethodName='by_visit_key')
    user_id = IntCol()


class Group(SQLObject):
    """
    An ultra-simple group definition.
    """
    # names like "Group", "Order" and "User" are reserved words in SQL
    # so we set the name to something safe for SQL
    class sqlmeta:
        table = 'tg_group'

    group_name = UnicodeCol(length=16, alternateID=True,
                            alternateMethodName='by_group_name')
    display_name = UnicodeCol(length=255)
    created = DateTimeCol(default=datetime.now)

    # collection of all users belonging to this group
    users = RelatedJoin('User', intermediateTable='user_group',
                        joinColumn='group_id', otherColumn='user_id')

    # collection of all permissions for this group
    permissions = RelatedJoin('Permission', joinColumn='group_id',
                              intermediateTable='group_permission',
                              otherColumn='permission_id')


class User(SQLObject):
    """
    Reasonably basic User definition.
    Probably would want additional attributes.
    """
    # names like "Group", "Order" and "User" are reserved words in SQL
    # so we set the name to something safe for SQL
    class sqlmeta:
        table = 'tg_user'

    user_name = UnicodeCol(length=16, alternateID=True,
                           alternateMethodName='by_user_name')
    email_address = UnicodeCol(length=255, alternateID=True,
                               alternateMethodName='by_email_address')
    display_name = UnicodeCol(length=255)
    password = UnicodeCol(length=40)
    created = DateTimeCol(default=datetime.now)

    # groups this user belongs to
    groups = RelatedJoin('Group', intermediateTable='user_group',
                         joinColumn='user_id', otherColumn='group_id')

    def _get_permissions(self):
        perms = set()
        for g in self.groups:
            perms = perms | set(g.permissions)
        return perms

    def _set_password(self, cleartext_password):
        "Runs cleartext_password through the hash algorithm before saving."
        password_hash = identity.encrypt_password(cleartext_password)
        self._SO_set_password(password_hash)

    def set_password_raw(self, password):
        "Saves the password as-is to the database."
        self._SO_set_password(password)


class Permission(SQLObject):
    """
    A relationship that determines what each Group can do
    """
    permission_name = UnicodeCol(length=16, alternateID=True,
                                 alternateMethodName='by_permission_name')
    description = UnicodeCol(length=255)

    groups = RelatedJoin('Group',
                         intermediateTable='group_permission',
                         joinColumn='permission_id',
                         otherColumn='group_id')
